''//////////////////////////////////////////////////////////////////////////////
'' Defender Remake                                                        
'' AUTHOR: Steve Waddicor
''//////////////////////////////////////////////////////////////////////////////

OBJ

  tv    : "sw_df_tv_drv_002.spin"
  gfx   : "sw_df_gfx_003.spin"
  rend  : "sw_df_gfx_renderer_003.spin"

CON
    _clkmode = xtal1 + pll8x
    _xinfreq = 10_000_000 + 0000
    _stack = 76
    COGS = 6     
    NUMBER_OF_SPRITES           = 20
    NUMBER_OF_LANDERS           = 10
    NUMBER_OF_LASERS            = 31

CON
    SPRITE_XPOS                 =0
    SPRITE_YPOS                 =1
    SPRITE_INVISIBLE            =%1000_0000_0000_0000

CON
    #0
    'states where user input accepted
    MARIO_FACING_LEFT
    MARIO_FACING_RIGHT
    MARIO_WALK_LEFT
    MARIO_WALK_RIGHT
    MARIO_DEAD_INIT
    MARIO_DEAD

CON
    #0
    GAME_TITLE_INIT
    GAME_TITLE
    GAME_START
    GAME_DISPLAY_PLAYER_INIT
    GAME_DISPLAY_PLAYER
    GAME_LEVEL_1_INIT
    GAME_LEVEL_1
    GAME_OVER_INIT
    GAME_OVER

VAR
    ' Renderer parameter block
    ' Each renderer copies these values to it's own cog_memory cache
    ' when it's got a little spare time, several times per frame
    ' so these values can be altered on the fly.
    long    rend_cog_number        'each cog gets a different number, starting with zero     write-only
    long    rend_scanline_req_adr  'a pointer to the request to put a line in the scan buffer   write-only
    long    rend_sprite_def_ptr    'pointer to tile definitions (longs)                      write-only
    long    rend_laser_def_ptr
    long    rend_palette_ptr       'pointer to colors (longs)                               write-only
    long    rend_scanline_buffer_adr 'hub memory where renderer puts scanline when requested write-only
    long    rend_cog_count         'total number of cogs                                     write-only
    long    rend_terrain_ptr
    long    rend_side_scroll

    long    rend_sprite_locs[NUMBER_OF_SPRITES]    '%i.YYYYYY_YYyyyyyy_XXXXXXXX_XXxxxxxx (negative = invisible)
    long    rend_sprite_tiles[NUMBER_OF_SPRITES]   '%EEEEEEEE_........_........_TTTTTTTT (E = explode count, T = tile number)
    long    rend_lasers[NUMBER_OF_LASERS]          '%........_........_NNNNNNNN_YYYYYYYY (N = state_count, Y = scan line.
                               
    'End of renderer parameter block

VAR
    ' TV parameter block
    long    tv_scanline_request
    long    tv_scanline_buffer
    long    tv_status
    ' End of TV parameter block

VAR
    long  startkey
    long  selectkey
    long  debug

VAR
    byte    game_state
    byte    player_level[2]
    byte    player_lives[2]
    byte    game_player_up
    byte    game_continue_life

    long    game_two_players
    long    player_score[2]
    long    hi_score
    long    gp
    long    game_state_count
    long    timer
    long    timer_state_count
    long    rnd

' New for Defender
    long    player_x_velocity
    long    player_y_velocity
    long    player_h_direction
    long    player_v_direction
    long    target_player_position
    long    player_position
    long    explode_count
    long    explode_state_count
    long    x_velocity[NUMBER_OF_SPRITES]
    long    y_velocity[NUMBER_OF_SPRITES]
    long    level_state_count
    long    laser_state_count
    long    a_button_up

    byte    dead[NUMBER_OF_SPRITES]
    

PUB Main | i, time, game_counter, total_time, avg_time, t
    'Mark top of stack to check later
    long[$7FFC]:=$DEAD_BEEF
    'Start renderers
    rend_sprite_def_ptr := gfx.sprite_adr
    rend_laser_def_ptr := gfx.laser_adr
    rend_palette_ptr := 0
    rend_scanline_buffer_adr := @hub_scanline_buffer
    rend_scanline_req_adr := @scanline_req
    rend_cog_count := COGS
    rend_terrain_ptr := gfx.terrain_adr
    SpriteClear

    repeat i from 0 to constant(cogs-1)
        rend_cog_number := i
        rend.start(@rend_cog_number)
        'Wait to allow this cog to get the cog number param from the block before changing it.
        repeat 1000

    tv_scanline_request := @scanline_req
    tv_scanline_buffer := @hub_scanline_buffer
    tv.start(@tv_scanline_request)

    'time:=CNT
    rnd:=$439845f0
    hi_score:=7650

    'Main Game Loop
    repeat
        'Check stack has not been overrun
        if long[$7FFC] <> $DEAD_BEEF
            DIRA[0]:=1
            OUTA[0]:=1
        'TV will let up know when there is a vsync
        tv_status := 0
        gp := NES_Read_Gamepad

        case game_state
            GAME_TITLE_INIT:
                SpriteClear
                game_state:=GAME_TITLE
            GAME_TITLE:
                if gp & GP_LEFT
                    game_two_players:=FALSE
                if gp & GP_RIGHT
                    game_two_players:=TRUE    
                'if gp & GP_START
                game_state:=GAME_START
            GAME_START:
                game_player_up := -game_two_players ' This is zero based.  Pretend player 2 was up last tyo make player 1 up this time.  
                game_state := GAME_LEVEL_1_INIT
            GAME_LEVEL_1_INIT:
                InitLevelCommon
                game_state := GAME_LEVEL_1
                rend_sprite_locs[0]:= $20_00_10_00
                rend_sprite_tiles[0]:= 0
                rnd := 419225
                player_h_direction:=1
                player_v_direction:=0
                repeat i from 1 to NUMBER_OF_LANDERS
                    dead[i] := TRUE
                repeat i from 0 to NUMBER_OF_LASERS
                    rend_lasers[i] := $ff
                rend_lasers[0] := 64
            GAME_LEVEL_1:
                PlayerLogic
                LanderLogic
                LaserLogic
'                DIRA[0]:=1
'                OUTA[0]:=1
            GAME_OVER_INIT:
                game_state:=GAME_OVER
            GAME_OVER:
                if game_state_count++ == 192
                    game_state:=GAME_TITLE_INIT

        repeat until tv_status>0
        'repeat 1000000

PRI SpriteClear | i
    repeat i from 0 to constant(NUMBER_OF_SPRITES-1)
        rend_sprite_locs[i]:= $8000_0000
        rend_sprite_tiles[i]:= 0

PRI InitLevelCommon
    SpriteClear

PRI PlayerLogic | i,advance

    if (gp & GP_LEFT)
        player_h_direction := -1
        rend_sprite_tiles[0] := 1
    elseif (gp & GP_RIGHT)
        player_h_direction := 1
        rend_sprite_tiles[0] := 0

    if (gp & GP_B)
        player_x_velocity+=player_h_direction*4
    if player_x_velocity>0
        player_x_velocity-=2
    elseif player_x_velocity<0
        player_x_velocity+=2

    if (gp & GP_A)
        if a_button_up
            rend_lasers[laser_state_count] :=rend_sprite_locs.word[SPRITE_YPOS]>>6
            a_button_up := FALSE
    else
        a_button_up := TRUE
 
    if (gp & GP_UP)
        if player_v_direction<>-1
            player_v_direction:=-1
            player_y_velocity:=0
        player_y_velocity-=6
    elseif (gp & GP_DOWN)
        if player_v_direction<>1
            player_v_direction:=1
            player_y_velocity:=0
        player_y_velocity+=6
    else
        player_y_velocity:=0

    player_x_velocity <#= 220
    player_x_velocity #>= -220
    player_y_velocity <#= 128
    player_y_velocity #>= -128
                      
    rend_sprite_locs.word[SPRITE_XPOS]+=player_x_velocity
    rend_sprite_locs.word[SPRITE_YPOS]+=player_y_velocity
    rend_sprite_locs.word[SPRITE_YPOS]<#=constant(221<<6)
    rend_sprite_locs.word[SPRITE_YPOS]#>=constant(39<<6)

    advance := ((player_x_velocity/6)-8)#>0
    'target_player_position
    target_player_position := 84-player_h_direction*48 +advance
    if player_position>target_player_position
        player_position-=1
    elseif player_position<target_player_position
        player_position+=1
    rend_side_scroll := ((rend_sprite_locs.word[SPRITE_XPOS]>>6)-player_position) & %1111111111

PRI LanderLogic | i, explode, this_lander, x, y, terrain
    level_state_count++
    if level_state_count == 400
        repeat i from 1 to NUMBER_OF_LANDERS/2
            rend_sprite_tiles[i]:= (45 << 16) + 2
            rend_sprite_locs[i]:= (45<<22) + (?rnd&$FFC0)
            dead[i] := FALSE
            x_velocity[i] := (?rnd&$F)+4
            if ?rnd&1
                x_velocity[i] := -x_velocity[i]    
     
    if level_state_count == 2000
        repeat i from NUMBER_OF_LANDERS/2+1 to NUMBER_OF_LANDERS
            rend_sprite_tiles[i]:= (45 << 16) + 2
            rend_sprite_locs[i]:= (45<<22) + (?rnd&$FFC0)
            dead[i] := FALSE
            x_velocity[i] := (?rnd&$F)+4
            if ?rnd&1
                x_velocity[i] := -x_velocity[i]    

    if explode_state_count==1
        repeat i from 1 to NUMBER_OF_LANDERS
            if dead[i] == FALSE
                explode := rend_sprite_tiles[i]>>16
                if explode<>0
                    explode--
                    rend_sprite_tiles[i] &= $FFFF
                    rend_sprite_tiles[i] |= (explode<<16)
        explode_state_count := 0
    else
        explode_state_count++
     
    repeat i from 1 to NUMBER_OF_LANDERS
        this_lander := @rend_sprite_locs[i] 
        word[this_lander][SPRITE_XPOS]+=x_velocity[i]
        
        x := word[this_lander][SPRITE_XPOS]>>6
        y := word[this_lander][SPRITE_YPOS]>>6
        terrain := byte[rend_terrain_ptr][x]
        if (terrain-y) > 40+i
            ' head down
            y_velocity[i]:=32
        elseif (terrain-y) < 16+i
            'head up
            y_velocity[i]:=-16
        else
            y_velocity[i]:=0
        word[this_lander][SPRITE_YPOS]+=y_velocity[i]

PRI LaserLogic | i,this_laser,tile
    this_laser:=@rend_lasers[0]
    repeat i from 0 to constant(NUMBER_OF_LASERS-1)
        tile := byte[this_laser][1]
        if tile==30
            byte[this_laser][0] := $FF
        else
            byte[this_laser][1] := ++tile
        this_laser += 4
    laser_state_count++
    if laser_state_count==31
        laser_state_count:=0
    
PRI Rotate(item_ptr,increment,lower,upper)
    ' Rotate a byte value through a range of values.  Useful for sprite animation.
    ' byte_item = pointer to a variable to ,reset,lower,upper)
    ' increment = value to add or subtract
    ' lower     = lower bound
    ' upper     = upper bound
    long[item_ptr]+=increment
    if long[item_ptr]>upper
        long[item_ptr] := lower
    if long[item_ptr]<lower
        long[item_ptr] := upper
    return

CON
' //////////////////////////////////////////////////////////////////
' NES Game Paddle Read
' //////////////////////////////////////////////////////////////////
' reads both gamepads in parallel encodes 8-bits for each in format
' right game pad #1 [15..8] : left game pad #0 [7..0]

    GP_RIGHT  = %00000001
    GP_LEFT   = %00000010
    GP_DOWN   = %00000100
    GP_UP     = %00001000
    GP_START  = %00010000
    GP_SELECT = %00100000
    GP_B      = %01000000
    GP_A      = %10000000

PUB NES_Read_Gamepad : nes_bits        |  i

' set I/O ports to proper direction
' P3 = JOY_CLK      (4)
' P4 = JOY_SH/LDn   (5)
' P5 = JOY_DATAOUT0 (6)
' P6 = JOY_DATAOUT1 (7)
' NES Bit Encoding

' step 1: set I/Os
DIRA [3] := 1 ' output
DIRA [4] := 1 ' output
DIRA [5] := 0 ' input
DIRA [6] := 0 ' input

' step 2: set clock and latch to 0
OUTA [3] := 0 ' JOY_CLK = 0
OUTA [4] := 0 ' JOY_SH/LDn = 0
'Delay(1)

' step 3: set latch to 1
OUTA [4] := 1 ' JOY_SH/LDn = 1
'Delay(1)

' step 4: set latch to 0
OUTA [4] := 0 ' JOY_SH/LDn = 0

' step 5: read first bit of each game pad

' data is now ready to shift out
' first bit is ready
nes_bits := 0

' left controller
nes_bits := INA[5] | (INA[6] << 8)

' step 7: read next 7 bits
repeat i from 0 to 6
 OUTA [3] := 1 ' JOY_CLK = 1
 'Delay(1)
 OUTA [3] := 0 ' JOY_CLK = 0
 nes_bits := (nes_bits << 1)
 nes_bits := nes_bits | INA[5] | (INA[6] << 8)

 'Delay(1)
' invert bits to make positive logic
nes_bits := (!nes_bits & $FFFF)

' //////////////////////////////////////////////////////////////////
' End NES Game Paddle Read
' //////////////////////////////////////////////////////////////////

DAT
' There seems to be some pointer leakage here.  Extra long as a patch.
                        long    0
' Hub scanline buffer = 256 pixels, 8 bits per pixel = 2048 bits = 256 bytes = 64 longs
hub_scanline_buffer
colorbar0               long    $0b_0b_0b_0b
colorbar1               long    $0c_0c_0c_0c
colorbar2               long    $0d_0d_0d_0d
colorbar3               long    $0e_0e_0e_0e
colorbar4               long    $1b_0b_1b_1b
colorbar5               long    $1c_0c_1c_1c
colorbar6               long    $1d_0d_1d_1d
colorbar7               long    $1e_0e_1e_1e
colorbar8               long    $2b_0b_2b_2b
colorbar9               long    $2c_2c_2c_2c
colorbar10              long    $2d_2d_2d_2d
colorbar11              long    $2e_2e_2e_2e
colorbar12              long    $3b_3b_3b_3b
colorbar13              long    $3c_3c_3c_3c
colorbar14              long    $3d_3d_3d_3d
colorbar15              long    $3e_3e_3e_3e
colorbar16              long    $4b_4b_4b_4b
colorbar17              long    $4c_4c_4c_4c
colorbar18              long    $4d_4d_4d_4d
colorbar19              long    $4e_4e_4e_4e
colorbar20              long    $5b_5b_5b_5b
colorbar21              long    $5c_5c_5c_5c
colorbar22              long    $5d_5d_5d_5d
colorbar23              long    $5e_5e_5e_5e
colorbar24              long    $6b_6b_6b_6b
colorbar25              long    $6c_6c_6c_6c
colorbar26              long    $6d_6d_6d_6d
colorbar27              long    $6e_6e_6e_6e
colorbar28              long    $7b_7b_7b_7b
colorbar29              long    $7c_7c_7c_7c
colorbar30              long    $7d_7d_7d_7d
colorbar31              long    $7e_7e_7e_7e
colorbar32              long    $8b_8b_8b_8b
colorbar33              long    $8c_8c_8c_8c
colorbar34              long    $8d_8d_8d_8d
colorbar35              long    $8e_8e_8e_8e
colorbar36              long    $9b_9b_9b_9b
colorbar37              long    $9c_9c_9c_9c
colorbar38              long    $9d_9d_9d_9d
colorbar39              long    $9e_9e_9e_9e
colorbar40              long    $ab_ab_ab_ab
colorbar41              long    $ac_ac_ac_ac
colorbar42              long    $ad_ad_ad_ad
colorbar43              long    $ae_ae_ae_ae
colorbar44              long    $bb_bb_bb_bb
colorbar45              long    $bc_bc_bc_bc
colorbar46              long    $bd_bd_bd_bd
colorbar47              long    $be_be_be_be
colorbar48              long    $cb_cb_cb_cb
colorbar49              long    $cc_cc_cc_cc
colorbar50              long    $cd_cd_cd_cd
colorbar51              long    $ce_ce_ce_ce
colorbar52              long    $db_db_db_db
colorbar53              long    $dc_dc_dc_dc
colorbar54              long    $dd_dd_dd_dd
colorbar55              long    $de_de_de_de
colorbar56              long    $eb_eb_eb_eb
colorbar57              long    $ec_ec_ec_ec
colorbar58              long    $ed_ed_ed_ed
colorbar59              long    $ee_ee_ee_ee
colorbar60              long    $fb_fb_fb_fb
colorbar61              long    $fc_fc_fc_fc
colorbar62              long    $fd_fd_fd_fd
colorbar63              long    $fe_fe_fe_fe
colorbar64              long    $0b_0b_0b_0b
colorbar65              long    $0c_0c_0c_0c
colorbar66              long    $0d_0d_0d_0d
colorbar67              long    $0e_0e_0e_0e
colorbar68              long    $1b_0b_1b_1b
colorbar69              long    $1c_0c_1c_1c
colorbar70              long    $1d_0d_1d_1d
colorbar71              long    $1e_0e_1e_1e
colorbar72              long    $2b_0b_2b_2b
colorbar73              long    $2c_2c_2c_2c
colorbar74              long    $2d_2d_2d_2d
colorbar75              long    $2e_2e_2e_2e
colorbar76              long    $3b_3b_3b_3b
colorbar77              long    $3c_3c_3c_3c
colorbar78              long    $3d_3d_3d_3d
colorbar79              long    $3e_3e_3e_3e
colorbar80              long    $4b_4b_4b_4b

' End of scanline buffer

scanline_req            long    0

 